import requests

from ..inning import Inning
from .server_fixture import server

def test_inning_raw_xml(server):
    gid = "gid_2018_03_27_anamlb_lanmlb_1"
    response = requests.get(server + "/inning.xml" )
    i = Inning(response.text, gid=gid)
    assert i.num == "1"
    assert i.next == "Y"
    assert i.away_team == "ana"
    assert i.gid == gid
    assert i.home_team == "lan"





from bs4.element import Tag as ELEMENT_TAG 
from bs4 import BeautifulSoup as BS


class Player(object):
  """
  Store player data.
  """

  IN_GAME_ATTRS = set(['bat_order', 'current_position', 'game_position'])
  PITCHER_ATTRS = set(['era', 'losses', 'wins'])
  COMMON_ATTRS = set([
     'avg',
     'bats',
     'boxname',
     'first',
     'hr',
     'id',
     'last',
     'num',
     'parent_team_abbrev',
     'parent_team_id',
     'position',
     'rbi',
     'rl',
     'status',
     'team_abbrev',
     'team_id'
  ])

  def __init__(self, xml=""):
      """
      Params
      ------
      xml : string | bs4.element.Tag instance
        Expects a single player tag.
      """
      player_tag = xml if type(xml) == ELEMENT_TAG else BS(xml, "lxml").find("player")

      self.in_game = False
      self.is_pitcher = False
      self.player_tag = player_tag

      for attr in player_tag.attrs:
          if attr in Player.COMMON_ATTRS:
              self.__dict__[attr] = player_tag[attr]
          elif attr in Player.PITCHER_ATTRS:
              self.is_pitcher = True 
              self.__dict__[attr] = player_tag[attr]
          elif attr in Player.IN_GAME_ATTRS:
              self.in_game = True
              self.__dict__[attr] = player_tag[attr]

  def get_id(self):             return int(self.id)
  def get_formatted_name(self): return self.first + " " + self.last




import pytest
import subprocess
import requests
import time 
import os

## start a simple http server that serves xml
@pytest.fixture(scope="module")
def server(request):
    import os
    port = 5555
    url = "http://localhost:{}".format(port)
    p = subprocess.Popen(
      ["python", "-m", "SimpleHTTPServer", str(port)], 
      stdout=subprocess.PIPE,
      cwd=find_xml_dir(os.getcwd()))

    attempts = 10
    while attempts > 0: 
        try:
            r = requests.get(url)
            attempts = 0
        except requests.exceptions.ConnectionError as e:
            attempts -= 1
            time.sleep(1/2.)

    def teardown():
        p.kill()

    request.addfinalizer(teardown)
    return url

def find_xml_dir(path):
    if os.path.basename(path) == 'xml':
        return path 

    for root, dirs, files in os.walk(path):
        if 'tests' in dirs:
            return find_xml_dir(os.path.join(path, 'tests'))
        elif 'hurlers' in dirs:
            return find_xml_dir(os.path.join(path, 'hurlers'))
        elif 'xml' in dirs:
            return find_xml_dir(os.path.join(path, 'xml'))



import sqlite3
import datetime
from bs4.element import Tag as ELEMENT_TAG 

class Pitch(object):
    """
    Store pitch data.
    """
    PITCH_ATTRS = set([
      'ax',
      'ay',
      'az',
      'batter', 
      'break_angle',
      'break_length',
      'break_y',
      'cc',
      'code',
      'des',
      ### ignore spanish for now 'des_es', 
      'end_speed',
      'end_tfs_zulu',
      'event',
      'event_num',
      'gid',
      'id',
      'inning',
      'mt',
      'nasty',
      'on_1b',
      'on_2b',
      'on_3b',
      'pfx_x',
      'pfx_z',
      'pitch_no',
      'pitch_type',
      'pitch_type_sequence',
      'pitcher',
      'play_guid',
      'px',
      'pz',
      'spin_dir',
      'spin_rate',
      'start_speed',
      'start_tfs_zulu',
      'sv_id',
      'sz_bot',
      'sz_top',
      'tfs',
      'tfs_zulu',
      'type',
      'type_confidence',
      'type_sequence',
      'vx0',
      'vy0',
      'vz0',
      'x',
      'x0',
      'y',
      'y0',
      'z0',
      'zone'
    ]) 
    IGNORE = set(['des_es'])

    INTEGER = set([
      'inning',
      'pitch_no',
      'id',
      'zone',
      'tfs',
      'event_num',
      'pitcher',
      'batter',
      'nasty',
      'on_1b',
      'on_2b',
      'on_3b'
    ] )

    REAL = set([
      'spin_rate',
      'break_angle',
      'sz_top',
      'y0',
      'ay',
      'ax',
      'az',
      'end_speed',
      'spin_dir',
      'start_speed',
      'pz',
      'px',
      'pfx_z',
      'vy0',
      'pfx_x',
      'break_length',
      'z0',
      'break_y',
      'vz0',
      'x0',
      'type_confidence',
      'y',
      'x',
      'sz_bot',
      'vx0'
    ])

    TEXT = set([
      'pitch_type_sequence',
      'code',
      'cc',
      'pitch_type',
      'pitch_id',
      'play_guid',
      'event',
      'gid',
      'type',
      'start_tfs_zulu',
      'tfs_zulu',
      'end_tfs_zulu',
      'type_sequence',
      'des',
      'mt',
      'sv_id'
    ])

    def __init__(self, xml, **kwargs):
        """
        Params
        ------
        xml : string | bs4.element.Tag instance 

        Keyword Arguments
        -----------------
        home_tag : bs4.element.Tag instance
          home team soup tag

        away_tag : bs4.element.Tag instance
          away team soup tag

        stadium_tag : bs4.element.Tag instance
          stadium soup tag

        Attribute Key
        ------------
        'x': Point for X(inches)
        'y': Point for Y(inches)
        'start_speed': The pitch speed(MPH) at the initial point
        'end_speed': The pitch speed(MPH) at the current batters
        'sz_top': The distance in feet from the ground to the top of the current batter's
        'sz_bot': The distance in feet from the ground to the bottom of the current batter's
        'pfx_x': The horizontal movement, in inches, of the pitch between the release point and home plate
        'pfx_z': The vertical movement, in inches, of the pitch between the release point and home plate
        'px': The left/right distance, in feet, of the pitch from the middle of the plate as it crossed home plate
        'pz': The height of the pitch in feet as it crossed the front of home plate
        'x0': The left/right distance, in feet, of the pitch, measured at the initial point
        'y0': The distance in feet from home plate where the PITCHf/x system is set to measure the initial parameters
        'z0': The height, in feet, of the pitch, measured at the initial point
        'vx0': The velocity of the pitch, in feet per second, in three dimensions, measured at the initial point
        'vy0': The velocity of the pitch, in feet per second, in three dimensions, measured at the initial point
        'vz0': The velocity of the pitch, in feet per second, in three dimensions, measured at the initial point
        'ax': The acceleration of the pitch, in feet per second per second, in three dimensions, measured at the initial point
        'ay': The acceleration of the pitch, in feet per second per second, in three dimensions, measured at the initial point
        'az': The acceleration of the pitch, in feet per second per second, in three dimensions, measured at the initial point
        'break_y': The distance in feet from the ground to the top of the current batter's
        'break_angle': The angle, in degrees, from vertical to the straight line path from the release point to where the pitch crossed the front of home plate, as seen from the catcher's/umpire's perspective
        'break_length': The measurement of the greatest distance, in inches, between the trajectory of the pitch at any point between the release point and the front of home plate
        'pitch_type': Pitch Type
        'pitch_type_seq': Pitch type Sequence, ex:FF|CU|FF
        'type_confidence': Pitch type confidence
        'zone': Pitch Zone
        'spin_dir': Pitch Spin Dir
        'spin_rate': Pitch Spin Rate
        'sv_id': Pitch in the air(From Datetime_To Datetime)
        'event_num': Event Sequence Number(atbat, pitch, action)
        """

        self.home_tag    = home_tag    if "home_tag"    in kwargs else None
        self.away_tag    = away_tag    if "away_tag"    in kwargs else None
        self.stadium_tag = stadium_tag if "stadium_tag" in kwargs else None

        ## presence varies from tag to tag
        self.on_1b = None
        self.on_2b = None
        self.on_3b = None

        ## instantiate all attrs
        for attr in Pitch.PITCH_ATTRS: self.__dict__[attr] = None

        ## store data from xml
        self._pitch_tag = xml if type(xml) == ELEMENT_TAG else BS(xml, "lxml").find("pitch")
        for attr in self._pitch_tag.attrs.keys():
            if attr in Pitch.PITCH_ATTRS:
                self.__dict__[attr] = self._pitch_tag.get(attr)
            else:
               if attr not in Pitch.IGNORE: 
                    print "Unregistered pitch attribute '{}'".format(attr)

        for attr in kwargs:
            if attr in Pitch.PITCH_ATTRS:
                self.__dict__[attr] = kwargs[attr]
            else:
                if attr not in Pitch.IGNORE: 
                    print "Unregistered pitch keyword '{}'".format(attr)


    @classmethod
    def create_pitch_id(cls, tfs_zulu, pitcher, id):
        import hashlib
        return hashlib.md5(tfs_zulu + str(pitcher) + str(id)).hexdigest()
           



from bs4.element import Tag as ELEMENT_TAG 
from bs4 import BeautifulSoup as BS
from .atbat import AtBat

class Inning(object):
    """
    Store inning data.
    """

    INN_ATTRS = set([ 
      "num",
      "next",
      "away_team",
      "gid",
      "home_team",
    ])
  
    def __init__(self, xml, **kwargs):
        """
        Params
        ------
        xml : string | bs4.element.Tag instance 
        """
        inning_tag = xml if type(xml) == ELEMENT_TAG else BS(xml, "lxml").find("inning")
        self.inning_tag = inning_tag
        for attr in inning_tag.attrs:
            if attr in Inning.INN_ATTRS:
                self.__dict__[attr] = inning_tag[attr]
            else:
                self.__dict__[attr] = inning_tag[attr]

        for attr in kwargs:
            if attr in Inning.INN_ATTRS:
                self.__dict__[attr] = kwargs[attr]
            else:
                self.__dict__[attr] = inning_tag[attr]

        top = inning_tag.find("top")
        bottom = inning_tag.find("bottom")

        self._top_atbats = map(lambda ab: AtBat(ab, gid=self.gid, inning=self.num), top.find_all("atbat")) if top else []
        self._bottom_atbats = map(lambda ab: AtBat(ab, gid=self.gid, inning=self.num), bottom.find_all("atbat")) if bottom else []
        self._atbats = self._top_atbats + self._bottom_atbats




from __future__ import print_function

import sys
import datetime
import requests
import re
from bs4 import BeautifulSoup as BS
from .constants import GID_P

def GET_page_soup(url):
    """
    Get and soupify the text at a url.    

    Params
    ------
    url : string
      Expects properly formatted URI.

    Note
    ----
    Methods that start with 'GET', make a network request.

    Returns
    -------
    If request is OK, returns soup of text. Else prints error code and returns
    `None`.
    """

    # grab the gameday id bits from the full url
    gid_match = re.search(GID_P, url)

    try: 
        response = requests.get(url, timeout=3)
    except (requests.ConnectionError, requests.Timeout) as e:
        ## FIXME not tested well
        if type(e) == requests.ConnectionError:
            print("[ERROR] Could not connect to '{}'".format(url), file=sys.stderr)
        else:
            print("[TIMEOUT] No response from '{}'.".format(url), file=sys.stderr)
        return

    status_is_ok = response.status_code >= 200 and response.status_code < 300
    if response.ok and status_is_ok:
        return BS(response.text, features="xml") 
    else:
        print("[ERROR] Bad response from {}. CODE: {}".format(url, str(response.status_code)), file=sys.stderr)
        return



import csv
from ..pitch import Pitch

def pitches_to_csv(filename, pitches):
    """
    Convert Pitch instances to a csv file.
    
    Params
    ------
    filename : string
      csv file
    pitches : Pitch instance
      pitches
    """
    csvfile = open(filename, "w")
    writer = csv.DictWriter(csvfile, sorted(Pitch.PITCH_ATTRS))
    writer.writeheader()
    for pitch in pitches:
        pitch_row_copy = { k:pitch.__dict__[k] for k in Pitch.PITCH_ATTRS }
        writer.writerow(pitch_row_copy)
    csvfile.close()
                



from __future__ import print_function

import sys
import os
import re
from bs4 import BeautifulSoup as BS
from .soup_utils import GET_page_soup
from .constants import *

def url_date_to_gameday(day):
    """
    Generate a gameday url from a datetime object.
    
    Params
    ------
    day : datetime.datetime
      year, month, and day required

    Returns
    -------
    String representing url for retrieving gameday xml.
    """
    try:
        year   = "year_{}".format(str(day.year))
        month  = "month_{}".format(str(day.month).zfill(2))
        day    = "day_{}".format(str(day.day).zfill(2))
    except AttributeError as e:
        print("[ERROR] Requires a datetime.datetime object. You passed a {}".format(type(day)), file=sys.stderr)
        return

    return MLBBASEURL +"/".join([year, month, day]) 

def url_to_gid(url):
    """
    Grab the gid from the gameday url. 
    
    Params
    ------
    url : str

    Returns
    -------
    String representing gameday id.
    """
    match = re.search(GID_P, url)
    if match:
        return url[match.start():match.end()]
    else:
        print("[ERROR] Could not find gid in '{}'".format(url), file=sys.stderr)
    return url


def url_gid_to_inning(gameday_url, inning="all", isfile=False):
    """
    Convert a gameday url to an inning url.

    Params
    ------
    gameday_url : string 
      (see Format in `GET_gid_urls`)
    inning : string
      1-9, all

    Returns
    -------
    Returns a url string to from which inning xml can be retrieved.
    """
    if isfile:
        return os.path.join(gameday_url, "inning_{}.xml".format(inning))
    else:
        return os.path.join(gameday_url, "inning", "inning_{}.xml".format(inning))

def url_gid_to_players(gameday_url, inning="all"):
    """
    Convert a gameday url to an inning url.

    Params
    ------
    gameday_url : string 
      (see Format in `GET_gid_urls`)
    inning : string
      1-9, all

    Returns
    -------
    Returns a url string to from which inning xml can be retrieved.
    """
    return os.path.join(gameday_url,"players.xml")

def url_gid_to_game(gid_url):
    """
    Convert a gid url to a game xml url

    Params
    ------
    gid_url : string 
      (see Format in `GET_gid_urls`)
    inning : string
      1-9, all

    Returns
    -------
    Returns a url string to from which inning xml can be retrieved.
    """
    return os.path.join(gid_url,"game.xml")

def soup_get_game_tags(game_soup):
    """
    Find the game links in a soup object.

    Params
    ------
    soup : bs4.BeautifulSoup object

    Returns
    -------
    A list of tags.

    """
    return game_soup.find_all('a', href=re.compile(GID_P))


def GET_gid_urls(day):
    """
    Retrieve the gameday urls. 

    Params
    ------
    day : datetime.datetime object

    Note
    ----
    Methods that start with GET, make a network request.

    Returns
    -------
    List of strings. Format `gid_{yyyy}_{mm}_{dd}_{awy}{lgu}_{hom}{lgu}_1`.
    'yyyy' : four digit year
    'mm' : two digit month
    'dd' : two digit day
    'awy': three letter abbreviation for away team
    'hom': three letter abbreviation for home team
    'lgu': three letter abbreviation for league

    """
    gameday_url = url_date_to_gameday(day)
    soup = GET_page_soup(gameday_url)
    if soup:
        gamehrefs = soup_get_game_tags(soup)
        return map(lambda tag: gameday_url + "/" + tag.contents[0].strip(), gamehrefs)   
    return []

## TODO clean this up and move to util package
def has_game(gid_url):
    GAME    = 0 ## 0x1
    PLAYERS = 0 ## 0x2
    INNING  = 0 ## 0x4
    soup = GET_page_soup(gid_url)
    for tag in soup.find_all('a'):
        href = tag.get('href')
        if re.match('.*game.xml$', href):    GAME    = 0x1 
        if re.match('.*players.xml$', href): PLAYERS = 0x2 
        if re.match('.*inning\/$', href):    INNING  = 0x4 
    return (GAME | PLAYERS | INNING) == 0x7




import requests

from bs4 import BeautifulSoup as BS

from ..runner import Runner
from .server_fixture import server


def test_runner_raw_xml(server):
    response = requests.get(server + "/runners.xml" )
    r = Runner(response.text)

    assert r.start == ""
    assert r.end == "1B"
    assert r.id == "645277"
    assert r.event == "Single"
    assert r.event_num == "10"

def test_runner_bs(server):
    response = requests.get(server + "/runners_bs.xml" )
    r = Runner(BS(response.text, "lxml").find("runner"))

    assert r.start == ""
    assert r.end == "1B"
    assert r.id == "645277"
    assert r.event == "Single"
    assert r.event_num == "10"



import time
import os
import datetime
from bs4 import BeautifulSoup as BS
from bs4.element import Tag as ELEMENT_TAG
from bs4 import BeautifulSoup as BS
from .utils.url_utils import *
from .utils.soup_utils import *
from .utils.csv_utils import *
from .inning import Inning
from .player import Player

class Game(object):
    """
    Get all the game data for a game id url.
    """

    GAME_ATTRS = set([
      "local_game_time",
      "game_pk",
      "game_time_et",
      "gameday_sw",
      "type"
    ])

    def __init__(self, gid_url):
        """
        Params
        ------
        gid_url : string
            gameday id url or gameday id directory

        """

        self._gid_url = gid_url
        self._gid = url_to_gid(gid_url)
        self._gidisdir = os.path.isdir(gid_url)
        if self._gidisdir: print "[XML ON DISK] stored @ {}".format(gid_url)

        ## convert gid to inning and player urls
        self._game_url = url_gid_to_game(gid_url)
        self._inning_all_url = url_gid_to_inning(gid_url, isfile=self._gidisdir)
        self._players_url = url_gid_to_players(gid_url)

        ## try to get the data over the network if the XML isn't on disk and
        ## soupify
        if self._gidisdir:
            with open(self._inning_all_url) as f:
                self._inning_all_soup = BS(f.read(), "lxml")
            with open(self._game_url) as f:
                self._game_soup = BS(f.read(), "lxml")
            with open(self._players_url) as f:
                self._players_soup = BS(f.read(), "lxml")
        else:
            self._inning_all_soup = GET_page_soup(self._inning_all_url)
            self._game_soup = GET_page_soup(self._game_url)
            self._players_soup = GET_page_soup(self._players_url)

        ## FIXME ugly
        self.successful_xml_retrieval = True

        ## unicode for printing to unicode enabled outputs
        telescope = u"\U0001f52d"
        thumbsup = u"\U0001f44d"
        try:
            print telescope +"  "+self._gid
        except:
            print "[PARSING] {}".format(self._gid)

        # mark true so client can check if Game instance is valid
        ## TODO (see line ~51)
        if not self._inning_all_soup or not self._game_soup or not self._players_soup:
            print "[ERROR] Did not store data for gid: '{}'.".format(self._gid)
            self.successful_xml_retrieval = False
            return

        ## TODO need to abort if none of the pages were retrieved

        ## add attributes to instance
        game_tag = self._game_soup.find("game")
        for attr in game_tag.attrs:
            if attr in Game.GAME_ATTRS:
                self.__dict__[attr] = game_tag[attr]
            else:
                print "[WARNING] Unregistered game attribute '{}'".format(attr)

        self._team_tags = game_tag.find_all("team")
        self._home_tag = self._team_tags[0] if self._team_tags[0].get("type") == "home" else self._team_tags[1]
        self._away_tag = self._team_tags[0] if self._team_tags[0].get("type") == "away" else self._team_tags[1]
        self._stadium_tag = game_tag.find("stadium")
        self._pitch_tags = self._inning_all_soup.find_all("pitch")

        # convert inning and player soup to Inning and Player instances,
        # respectively
        if not self._inning_all_soup:
            print "[WARNING] no inning soup"
        self._innings = map(lambda i: Inning(i, gid=self._gid), self._inning_all_soup.find_all("inning"))
        self._players = map(Player, self._players_soup.find_all("player"))

        ## map ids to player instances
        self._player_map = {}
        for player in self._players:
          self._player_map[player.id] = player

        ## cache pitcher ids
        self._pitcher_ids = set()

        ## cache all pitches in game
        self._pitches = []

        try:
            print thumbsup + "  " + self.__str__() + "\n"
        except:
            print "[SUCCESS]" + "  " + self.__str__() + "\n"

    def __str__(self):
        return "{} vs {}".format(self._away_tag.get("name_full"), self._home_tag.get("name_full"))

    def _get_player_name(self, id):
        """
        Convert player id to '[first name] [last name]'.
        """
        id = str(id) if type(id) == int else id
        p = self._player_map[id]
        return p.first + " " + p.last

    def _store_pitcher_ids(self):
        for inning in self._innings:
            for atbat in inning._atbats:
                if atbat.pitcher not in self._pitcher_ids:
                    self._pitcher_ids.add(atbat.pitcher)

    def get_pitches(self):
        """
        All pitches in game.

        Returns
        -------
        list of Pitch instances
        """
        if not self._pitches:
            self._pitches = [pitch for inning in self._innings for atbat in inning._atbats for pitch in atbat._pitches]
        return self._pitches

    def get_pitchers(self):
        """
        Participating pitchers. Not in order.

        Returns
        -------
        list of Player instances.
        """
        if not self._pitcher_ids:
            self._store_pitcher_ids()
        return filter(lambda p: p.id in self._pitcher_ids, self._players)

    def get_pitcher_names(self):
        """
        Participating pitchers. Not in order.

        Returns
        -------
        list of strings.

        """
        if not self._pitcher_ids:
            self._store_pitcher_ids()
        return map(lambda id: "{} - ".format(id) + self._get_player_name(id), self._pitcher_ids)

    def print_pitchers(self):
        for name in self.get_pitcher_names():
            print name



from bs4.element import Tag as ELEMENT_TAG 
from bs4 import BeautifulSoup as BS

class Runner(object):
    """
    Store runner data.
    """

    RUNNER_ATTRS = {
      'earned',
      'end',
      'event',
      'event_num',
      'id',
      'rbi',
      'score',
      'start'
    }
    def __init__(self, xml):
       """
       Params
       ---------
       xml : string | bs4.element.Tag
       """
       runner_tag = xml if type(xml) == ELEMENT_TAG else BS(xml, "lxml").find("runner")
       self.runner_tag = runner_tag
       for attr in runner_tag.attrs:
           if attr in Runner.RUNNER_ATTRS:
               self.__dict__[attr] = runner_tag[attr]
           else:
               print "[WARNING] Unregistered runner attribute '{}'".format(attr)
               self.__dict__[attr] = runner_tag[attr]

    



import requests

from ..game import Game

from .server_fixture import server

def test_game(server):
    gid = "gid_2018_10_28_bosmlb_lanmlb_1"
    g = Game(server + "/" + gid)

    pitcher_ids = set([
        '523260',
        '445276',
        '477132',
        '519242',
        '456034',
        '520980'
    ])

    pitchers = g.get_pitchers()

    assert len(pitchers) == len(pitcher_ids) 
    for pitcher in pitchers:
        assert pitcher.id in pitcher_ids



import requests

from ..atbat import AtBat
from .server_fixture import server

def test_atbat_raw_xml(server):
    gid = "gid_2018_03_27_anamlb_lanmlb_1"
    inning = 1
    response = requests.get(server + "/atbat.xml" )
    a = AtBat(response.text, gid=gid, inning=1)

    assert a.gid == gid
    assert a.inning == 1 
    assert a.away_team_runs == "0"
    assert a.b == "2"
    assert a.b_height == "6-0"
    assert a.batter == "446359"
    assert a.des == "Zack Cozart lines out sharply to left fielder Matt Kemp.  " 
    assert a.end_tfs_zulu == "2018-03-28T02:11:52Z" 
    assert a.event == "Lineout" 
    assert a.event_num == "10" 
    assert a.home_team_runs == "0" 
    assert a.num == "1" 
    assert a.o == "1" 
    assert a.p_throws == "L" 
    assert a.pitcher == "547943" 
    assert a.play_guid == "3ee5fe7a-29fc-421e-adec-5629aaa796c3" 
    assert a.s == "2" 
    assert a.stand == "R" 
    assert a.start_tfs == "030915" 
    assert a.start_tfs_zulu == "2018-03-28T03:09:15Z"

def test_atbat_raw_xml_pitch_type_sequence(server):
    gid = "gid_2018_03_27_anamlb_lanmlb_1"
    inning = 1
    response = requests.get(server + "/atbat.xml" )
    a = AtBat(response.text, gid=gid, inning=1)
    assert a._pitch_type_sequence == ['FF', 'FC', 'CU', 'FF', 'CU', 'FF']



from setuptools import setup, find_packages


setup(
    name='hurlers',
    version='0.1.1',
    description='Store mlbam pitch data.',
    author='S.J. Browne',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages(),
    include_package_data=True,
    keywords=['baseball', 'MLB', 'MLBAM', 'SABRmetrics', 'SABR', 'Major League Baseball', 'pitch', 'pitching'],
    license='MIT License',
    install_requires=[
        'bs4',
        'lxml',
        'requests',
    ],
    url='https://github.com/sjbrowne/hurlers',
    python_requires=">=2.7"
)



from bs4.element import Tag as ELEMENT_TAG 
from bs4 import BeautifulSoup as BS
from .pitch import Pitch
from .runner import Runner

class AtBat(object):
    """
    Store atbat data.
    """
    AB_ATTRS = set([
      'away_team_runs',
      'b',
      'b_height',
      'batter',
      'des',
      'des_es',
      'earned', 
      'end_tfs_zulu',
      'event',
      'event2',
      'event_es',
      'event_num',
      'gid',
      'home_team_runs',
      'inning',
      'num',
      'o',
      'p_throws',
      'pitcher',
      'play_guid',
      'rbi',
      's',
      'score',
      'stand',
      'start_tfs',
      'start_tfs_zulu'
    ])

    IGNORE = set()

    def __init__(self, xml, **kwargs):
        """
        Params
        ------
        xml : string | bs4.element.Tag instance 
        """

        ## soupify if not soupified
        atbat_tag = xml if type(xml) == ELEMENT_TAG else BS(xml, "lxml").find("atbat")
        self.atbat_tag = atbat_tag

        ## store each attribute on the 'atBat' tag
        for attr in atbat_tag.attrs:
            if attr in AtBat.AB_ATTRS:
                self.__dict__[attr] = atbat_tag[attr]
            else:
                self.__dict__[attr] = atbat_tag[attr]

        ## store user defined data
        for attr in kwargs:
            if attr in AtBat.AB_ATTRS:
                self.__dict__[attr] = kwargs[attr]
            else:
                if attr not in AtBat.IGNORE:
                    self.__dict__[attr] = kwargs[attr]
        
        ## map 'runner' tags to Runner instances 
        self._runners = map(Runner, atbat_tag.find_all("runner"))

        ## preprocessing for storing pitches
        self._pitch_tags = atbat_tag.find_all("pitch") 
        self._store_pitch_type_sequence_and_type_sequence()
        keywords = {
            'pitcher': self.pitcher,
            'batter':  self.batter,
            'start_tfs_zulu': self.start_tfs_zulu,
            'end_tfs_zulu': self.end_tfs_zulu,
            'inning': self.inning,
            'gid': self.gid
        }

        ## map 'pitch' tags to Pitch instances 
        self._pitches = map(lambda t: Pitch(t, **keywords), self._gen_pitches())

    ## FIXME adding attributes to pitches this way is hacky but fine for now; it
    ## is assumed that most features will be developed post processing, but if
    ## that changes, adding features to pitches based on existing features should
    ## be easier/cleaner
    def _gen_pitches(self):
        """
        Generator of extended 'pitch' tags. 
        """
        i = 0 
        pitches = self.atbat_tag.find_all("pitches") 
        while i < len(self._pitch_tags):
            p = self._pitch_tags[i]

            p.attrs['pitch_no'] = i+1
            p.attrs['type_sequence'] = self._type_sequence
            p.attrs['pitch_type_sequence'] = self._pitch_type_sequence 
            p.attrs['event'] = self.atbat_tag.get("event") 
            yield p
            i+=1

    ## REQUIRES self._pitch_tags to be defined
    def _store_pitch_type_sequence_and_type_sequence(self):
        self._pitch_type_sequence, self._type_sequence = [], []
        for p in self._pitch_tags:
            self._pitch_type_sequence.append(p.get("pitch_type"))
            self._type_sequence.append(p.get("type"))
            

